"""Autogenerated DynamoDB exceptions.

This file was autogenerated by scripts/generate_exceptions.py.
Do not edit it manually!

"""
from dokklib_db_extended.errors.client import ClientError


class BackupInUseException(ClientError):
    """Please check DynamoDB docs for documentation."""


class BackupNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ConditionalCheckFailedException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ContinuousBackupsUnavailableException(ClientError):
    """Please check DynamoDB docs for documentation."""


class GlobalTableAlreadyExistsException(ClientError):
    """Please check DynamoDB docs for documentation."""


class GlobalTableNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class IdempotentParameterMismatchException(ClientError):
    """Please check DynamoDB docs for documentation."""


class IndexNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class InternalServerError(ClientError):
    """Please check DynamoDB docs for documentation."""


class InvalidRestoreTimeException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ItemCollectionSizeLimitExceededException(ClientError):
    """Please check DynamoDB docs for documentation."""


class LimitExceededException(ClientError):
    """Please check DynamoDB docs for documentation."""


class PointInTimeRecoveryUnavailableException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ProvisionedThroughputExceededException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ReplicaAlreadyExistsException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ReplicaNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class RequestLimitExceeded(ClientError):
    """Please check DynamoDB docs for documentation."""


class ResourceInUseException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ResourceNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class TableAlreadyExistsException(ClientError):
    """Please check DynamoDB docs for documentation."""


class TableInUseException(ClientError):
    """Please check DynamoDB docs for documentation."""


class TableNotFoundException(ClientError):
    """Please check DynamoDB docs for documentation."""


class TransactionConflictException(ClientError):
    """Please check DynamoDB docs for documentation."""


class TransactionInProgressException(ClientError):
    """Please check DynamoDB docs for documentation."""


class ValidationError(ClientError):
    """Please check DynamoDB docs for documentation."""


class ThrottlingError(ClientError):
    """Please check DynamoDB docs for documentation."""
